unit BlendFill;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
 TBlendFillType = (bftHorz,bftVert,bftCircular,bftElliptic,
                   bftSquare,bftRect,bftConic,bftAngle,bftQuad);

type
 TRegion = (rgnRectangle,rgnElliptic);

type
  TBlendFill = class(TGraphicControl)
  private
   FBlendFillType  : TBlendFillType;
   FColor1,FColor2,
   FColor3,FColor4 : TColor;
   FColBands       : Byte;
   FAngle          : Integer;
   FRegion         : TRegion;
   procedure SetBlendFillType(aValue : TBlendFillType);
   procedure SetColor1(aValue : TColor);
   procedure SetColor2(aValue : TColor);
   procedure SetColor3(aValue : TColor);
   procedure SetColor4(aValue : TColor);
   procedure SetColBands(aValue : Byte);
   procedure SetAngle(aValue : Integer);
   procedure SetRegion(aValue : TRegion);
  protected
   procedure Paint; override;
   procedure WMPaletteChanged(var Msg : TWMPaletteChanged); message wm_PaletteChanged;
  public
   constructor Create(aOwner : TComponent); override;
  published
   (* geerbte properties *)
   property Align;
   property DragCursor;
   property DragMode;
   property Enabled;
   property ParentShowHint;
   property PopupMenu;
   property ShowHint;
   property Visible;
   property OnClick;
   property OnDblClick;
   property OnDragDrop;
   property OnEndDrag;
   property OnMouseDown;
   property OnMouseMove;
   property OnMouseUp;
   property OnStartDrag;
   (* eigene properties *)
   property BlendFillType : TBlendFillType read FBlendFillType write SetBlendFillType;
   property Color1 : TColor read FColor1 write SetColor1;
   property Color2 : TColor read FColor2 write SetColor2;
   property Color3 : TColor read FColor3 write SetColor3;
   property Color4 : TColor read FColor4 write SetColor4;
   property ColBands : Byte read FColBands write SetColBands;
   property Angle : Integer read FAngle write SetAngle;
   property Region : TRegion read FRegion write SetRegion;
  end;

procedure Register;

implementation

uses
 BlendFillProcs;

procedure TBlendFill.SetBlendFillType;
begin
 if (FBlendFillType<>aValue) then begin
  FBlendFillType:=aValue;
  Invalidate;
 end;
end;

procedure TBlendFill.SetColor1;
begin
 if (Color1<>aValue) then begin
  FColor1:=aValue;
  Invalidate;
 end;
end;

procedure TBlendFill.SetColor2;
begin
 if (Color2<>aValue) then begin
  FColor2:=aValue;
  Invalidate;
 end;
end;

procedure TBlendFill.SetColor3;
begin
 if (Color3<>aValue) then begin
  FColor3:=aValue;
  Invalidate;
 end;
end;

procedure TBlendFill.SetColor4;
begin
 if (Color4<>aValue) then begin
  FColor4:=aValue;
  Invalidate;
 end;
end;

procedure TBlendFill.SetColBands;
begin
 if (FColBands<>aValue) then begin
  FColBands:=aValue;
  Invalidate;
 end;
end;

procedure TBlendFill.SetAngle;
begin
 if (aValue<-180) then aValue:=-180;
 if (aValue>180) then aValue:=180;
 if (FAngle<>aValue) then begin
  FAngle:=aValue;
  if (FBlendFillType=bftAngle) then Invalidate;
 end;
end;

procedure TBlendFill.SetRegion;
begin
 if (FRegion<>aValue) then begin
  FRegion:=aValue;
  Invalidate;
 end;
end;

procedure TBlendFill.Paint;
var
 Rgn  : HRgn;
 Rect : TRect;
begin
 if csDesigning in ComponentState then with Canvas do begin
  Pen.Style:=psDash;
  Brush.Style:=bsClear;
  Rectangle(0,0,Width,Height);
 end;
 Rgn:=0;
 case FRegion of
  rgnRectangle:
   Rgn:=CreateRectRgn(Left,Top,Left+Width,Top+Height);
  rgnElliptic:
   Rgn:=CreateEllipticRgn(Left,Top,Left+Width,Top+Height);
 end;
 SetRect(Rect,0,0,Width,Height);
 case FBlendFillType of
  bftHorz:
   LinearBlendFill(Canvas.Handle,Rgn,Rect,FColor1,FColor2,lbfHorz,FColBands);
  bftVert:
   LinearBlendFill(Canvas.Handle,Rgn,Rect,FColor1,FColor2,lbfVert,FColBands);
  bftCircular:
   RadialBlendFill(Canvas.Handle,Rgn,Rect,FColor1,FColor2,rbfCircular,FColBands);
  bftElliptic:
   RadialBlendFill(Canvas.Handle,Rgn,Rect,FColor1,FColor2,rbfElliptic,FColBands);
  bftSquare:
   EdgeBlendFill(Canvas.Handle,Rgn,Rect,FColor1,FColor2,ebfSquare,FColBands);
  bftRect:
   EdgeBlendFill(Canvas.Handle,Rgn,Rect,FColor1,FColor2,ebfRect,FColBands);
  bftConic:
   ConicBlendFill(Canvas.Handle,Rgn,Rect,FColor1,FColor2,FColBands);
  bftAngle:
   AngleBlendFill(Canvas.Handle,Rgn,Rect,FColor1,FColor2,FAngle,FColBands);
  bftQuad:
   QuadBlendFill(Canvas.Handle,Rgn,Rect,FColor1,FColor2,FColor3,FColor4,FColBands,FColBands);
 end;
 DeleteObject(Rgn);
end;

procedure TBlendFill.WMPaletteChanged;
begin
 Invalidate;
end;

constructor TBlendFill.Create;
begin
 inherited Create(aOwner);
 Width:=300;
 Height:=200;
 FBlendFillType:=bftHorz;
 FColor1:=clBlue;
 FColor2:=clBlack;
 FColor3:=clWhite;
 FColor4:=clRed;
 FColBands:=100;
 FAngle:=45;
 FRegion:=rgnRectangle;
end;

procedure Register;
begin
 RegisterComponents('CDBUtil',[TBlendFill]);                                    //wgg was exemples
end;

end.
