unit BlendFillProcs;

interface

uses
 WinTypes,WinProcs,Graphics;

type
 (* Sub-Typen der Verlufe *)
 TLinearBlendFill = (lbfVert,lbfHorz);
 TRadialBlendFill = (rbfCircular,rbfElliptic);
 TEdgeBlendFill   = (ebfSquare,ebfRect);

(* linear *)
procedure LinearBlendFill(DC : THandle;Rgn : HRgn;Rect : TRect;Col1,Col2 : TColor;
          FillType : TLinearBlendFill;ColBands : Byte);
(* kreisfrmig *)
procedure RadialBlendFill(DC : THandle;Rgn : HRgn;Rect : TRect;Col1,Col2 : TColor;
          FillType : TRadialBlendFill;ColBands : Byte);
(* eckig *)
procedure EdgeBlendFill(DC : THandle;Rgn : HRgn;Rect : TRect;Col1,Col2 : TColor;
          FillType : TEdgeBlendFill;ColBands : Byte);
(* konisch *)
procedure ConicBlendFill(DC : THandle;Rgn : HRgn;Rect : TRect;Col1,Col2 : TColor;
          ColBands : Byte);
(* schrg *)
procedure AngleBlendFill(DC : THandle;Rgn : HRgn;Rect : TRect;Col1,Col2 : TColor;
          Angle : Integer;ColBands : Byte);
(* 4 Farben *)
procedure QuadBlendFill(DC : THandle;Rgn : HRgn;Rect : TRect;Col1,Col2,Col3,Col4 : TColor;
          xTile,yTile : Longint);

implementation

const
 minColBands = 3;    (* minimale Schritte *)
 maxColBands = 255;  (* maximale Schritte *)

type
 (* fr vorberechnete Farben *)
 TBlendCol = array[0..maxColBands]of TColor;

(* korrigiert unsinnige Eingaben *)
function CorrectColBands(ColBands : Byte) : Byte;
begin
 if (ColBands<minColBands) then Result:=minColBands
// else if (ColBands>maxColBands) then Result:=maxColBands
 else Result:=ColBands;
end;

(* berechnet die Farben *)
procedure PrepareColors(Col1,Col2 : TColor;ColBands : Byte;var BlendCol : TBlendCol);
var
 dRGB : array[1..3]of Integer;
 z    : Integer;
begin
 dRGB[1]:=GetRValue(Col2)-GetRValue(Col1);
 dRGB[2]:=GetGValue(Col2)-GetGValue(Col1);
 dRGB[3]:=GetBValue(Col2)-GetBValue(Col1);
 BlendCol[0]:=RGB(GetRValue(Col1),GetGValue(Col1),GetBValue(Col1));
 for z:=1 to ColBands do begin
  BlendCol[z]:=RGB(Round(GetRValue(BlendCol[0])+z*(dRGB[1]/ColBands)),
                   Round(GetGValue(BlendCol[0])+z*(dRGB[2]/ColBands)),
                   Round(GetBValue(BlendCol[0])+z*(dRGB[3]/ColBands)));
 end;
end;

(* bereiten das umschlieende Recteck vor *)
{procedure PrepareRgnBox(Rgn : HRgn;var Box : TRect);
begin
 GetRgnBox(Rgn,Box);
 Box.Bottom:=Box.Bottom-Box.Top;
 Box.Right:=Box.Right-Box.Left;
end;
}
(* wandelt einen Winkel in Rad um *)
function Rad(Angle : Double) : Double;
begin
 Angle:=Angle/360;
 Angle:=Angle-Int(Angle);
 Result:=2*Pi*Angle;
end;

(* linear *)
procedure LinearBlendFill;
var
 RClip,RPat : TRect;
 Brush      : HBrush;
 z          : Integer;
 BlendCol   : TBlendCol;
begin
 (* initialisieren *)
 SelectClipRgn(DC,Rgn);
 RClip:=Rect;
 ColBands:=CorrectColBands(ColBands);
 PrepareColors(Col1,Col2,ColBands,BlendCol);
 RPat:=RClip;
 for z:=0 to ColBands do begin
  case FillType of
   lbfVert:
   begin
    (* vertikal *)
    RPat.Left:=Round(RClip.Left+z*(RClip.Right/ColBands));
    RPat.Right:=Round(RClip.Left+RClip.Right/ColBands)+1;
   end;
   lbfHorz:
   begin
    (* horizontal *)
    RPat.Top:=Round(RClip.Top+z*(RClip.Bottom/ColBands));
    RPat.Bottom:=Round(RClip.Top+RClip.Bottom/ColBands)+1;
   end;
  end;
  Brush:=SelectObject(DC,CreateSolidBrush(BlendCol[z]));
  PatBlt(DC,RPat.Left,RPat.Top,RPat.Right,RPat.Bottom,PatCopy);
  DeleteObject(SelectObject(DC,Brush));
 end;
 SelectClipRgn(DC,0);
end;


(* kreisfrmig *)
procedure RadialBlendFill;
var
 RClip,RPat : TRect;
 Brush      : HBrush;
 Pen        : HPen;
 z,d        : Integer;
 BlendCol   : TBlendCol;
 MemDC      : HDC;
 Bitmap     : HBitmap;
begin
 (* initialisieren *)
 SelectClipRgn(DC,Rgn);
 RClip:=Rect;
 MemDC:=CreateCompatibleDC(DC);
 Bitmap:=SelectObject(MemDC,CreateCompatibleBitmap(DC,RClip.right,RClip.bottom));
 ColBands:=CorrectColBands(ColBands);
 PrepareColors(Col1,Col2,ColBands,BlendCol);
 Pen:=SelectObject(MemDC,GetStockObject(Null_Pen));
 Brush:=CreateSolidBrush(Col1);
 SelectObject(MemDC,Brush);
 Rectangle(MemDC,0,0,RClip.Right+1,RClip.Bottom+1);
 DeleteObject(Brush);
 for z:=0 to ColBands do begin
  case FillType of
   rbfCircular:
   begin
    (* kreisfrmig *)
    if (RClip.Bottom>RClip.Right) then d:=RClip.Bottom
                                  else d:=RClip.Right;
    RPat.Top:=Round(RClip.Top+RClip.Bottom/2-(ColBands-z)*(d/ColBands)/2);
    RPat.Bottom:=Round(RClip.Top+RClip.Bottom/2+(ColBands-z)*(d/ColBands)/2);
    RPat.Left:=Round(RClip.Left+RClip.Right/2-(ColBands-z)*(d/ColBands)/2);
    RPat.Right:=Round(RClip.Left+RClip.Right/2+(ColBands-z)*(d/ColBands)/2);
   end;
   rbfElliptic:
   begin
    (* elliptisch *)
    RPat.Top:=Round(RClip.Top+z*(RClip.Bottom/ColBands)/2);
    RPat.Bottom:=Round(RClip.Top+RClip.Bottom-z*(RClip.Bottom/ColBands)/2);
    RPat.Left:=Round(RClip.Left+z*(RClip.Right/ColBands)/2);
    RPat.Right:=Round(RClip.Left+RClip.Right-z*(RClip.Right/ColBands)/2);
   end;
  end;
  Brush:=SelectObject(MemDC,CreateSolidBrush(BlendCol[z]));
  Ellipse(MemDC,RPat.Left,RPat.Top,RPat.Right,RPat.Bottom);
  DeleteObject(SelectObject(MemDC,Brush));
 end;
 BitBlt(DC,RClip.left,RClip.top,RClip.right,RClip.bottom,MemDC,RClip.left,RClip.top,SrcCopy);
 SelectClipRgn(DC,0);
 DeleteObject(SelectObject(MemDC,Pen));
 DeleteObject(SelectObject(MemDC,Bitmap));
 DeleteDC(MemDC);
end;

(* eckig *)
procedure EdgeBlendFill;
var
 RClip,RPat : TRect;
 Brush      : HBrush;
 z,d        : Integer;
 BlendCol   : TBlendCol;
 MemDC      : HDC;
 Bitmap     : HBitmap;
begin
 (* initialisieren *)
 SelectClipRgn(DC,Rgn);
 RClip:=Rect;
 MemDC:=CreateCompatibleDC(DC);
 Bitmap:=SelectObject(MemDC,CreateCompatibleBitmap(DC,RClip.right,RClip.bottom));
 ColBands:=CorrectColBands(ColBands);
 PrepareColors(Col1,Col2,ColBands,BlendCol);
 for z:=0 to ColBands do begin
  case FillType of
   ebfSquare:
   begin
    (* quadratisch *)
    if (RClip.Bottom>RClip.Right) then d:=RClip.Bottom
                                  else d:=RClip.Right;
    RPat.Top:=Round(RClip.Top+RClip.Bottom/2-(ColBands-z)*(d/ColBands)/2);
    RPat.Bottom:=Round(RClip.Top+RClip.Bottom/2+(ColBands-z)*(d/ColBands)/2);
    RPat.Left:=Round(RClip.Left+RClip.Right/2-(ColBands-z)*(d/ColBands)/2);
    RPat.Right:=Round(RClip.Left+RClip.Right/2+(ColBands-z)*(d/ColBands)/2);
   end;
   ebfRect:
   begin
    (* rechteckig *)
    RPat.Top:=Round(RClip.Top+z*(RClip.Bottom/ColBands)/2);
    RPat.Bottom:=Round(RClip.Top+RClip.Bottom-z*(RClip.Bottom/ColBands)/2);
    RPat.Left:=Round(RClip.Left+z*(RClip.Right/ColBands)/2);
    RPat.Right:=Round(RClip.Left+RClip.Right-z*(RClip.Right/ColBands)/2);
   end;
  end;
  Brush:=SelectObject(MemDC,CreateSolidBrush(BlendCol[z]));
  PatBlt(MemDC,RPat.Left,RPat.Top,RPat.Right-RPat.Left,RPat.Bottom-RPat.Top,PatCopy);
  DeleteObject(SelectObject(MemDC,Brush));
 end;
 BitBlt(DC,RClip.left,RClip.top,RClip.right,RClip.bottom,MemDC,RClip.left,RClip.top,SrcCopy);
 SelectClipRgn(DC,0);
 DeleteObject(SelectObject(MemDC,Bitmap));
 DeleteDC(MemDC);
end;

(* konisch *)
procedure ConicBlendFill;
var
 RClip       : TRect;
 PPatT,PPatB : array[0..2]of TPoint; (* fr Dreiecke *)
 Brush       : HBrush;
 Pen         : HPen;
 z,d         : Integer;
 a           : Double;
 BlendCol    : TBlendCol;
 MemDC       : HDC;
 Bitmap      : HBitmap;
begin
 (* initialisieren *)
 SelectClipRgn(DC,Rgn);
 RClip:=Rect;
 MemDC:=CreateCompatibleDC(DC);
 Bitmap:=SelectObject(MemDC,CreateCompatibleBitmap(DC,RClip.right,RClip.bottom));
 ColBands:=CorrectColBands(ColBands);
 PrepareColors(Col1,Col2,ColBands,BlendCol);
 if (RClip.Bottom>RClip.Right) then d:=RClip.Bottom
                               else d:=RClip.Right;
 (* Durchmesser des Kreises *)
 d:=Round(Sqrt(2*Sqr(d)));
 (* Mittelpunkt (fr alle Dreiecke gleich) *)
 PPatT[0].X:=Round(RClip.Left+RClip.Right/2);
 PPatT[0].Y:=Round(RClip.Top+RClip.Bottom/2);
 PPatB[0]:=PPatT[0];
 Pen:=SelectObject(MemDC,GetStockObject(Null_Pen));
 (* Startwerte oben *)
 PPatT[1].X:=Round(PPatT[0].X+d/2);
 PPatT[1].Y:=PPatT[0].Y;
 (* Startwinkel *)
 a:=Pi/ColBands;
 (* Startwerte unten *)
 PPatB[1]:=PPatT[1];
 for z:=0 to ColBands do begin
  Brush:=SelectObject(MemDC,CreateSolidBrush(BlendCol[z]));
  (* oben *)
  a:=a-Pi/ColBands;
  PPatT[2].X:=PPatT[0].X+Round(Cos(a)*d);
  PPatT[2].Y:=PPatT[0].Y+Round(Sin(a)*d);
  Polygon(MemDC,PPatT,3);
  PPatT[1]:=PPatT[2];
  (* unten *)
  PPatB[2].X:=PPatT[2].X;
  PPatB[2].Y:=PPatT[2].Y+2*(PPatB[0].Y-PPatT[2].Y);
  Polygon(MemDC,PPatB,3);
  PPatB[1]:=PPatB[2];
  DeleteObject(SelectObject(MemDC,Brush));
 end;
 DeleteObject(SelectObject(MemDC,Pen));
 BitBlt(DC,RClip.left,RClip.top,RClip.right,RClip.bottom,MemDC,RClip.left,RClip.top,SrcCopy);
 SelectClipRgn(DC,0);
 DeleteObject(SelectObject(MemDC,Bitmap));
 DeleteDC(MemDC);
end;

(* schrg *)
procedure AngleBlendFill;
var
 RClip         : TRect;
 mx,my,z       : Integer;
 Pat           : array[0..3]of TPoint;
 Brush         : HBrush;
 Pen           : HPen;
 sx,sy,d       : Double;
 Start0,Start1 : TPoint;
 BlendCol      : TBlendCol;
 MemDC         : HDC;
 Bitmap        : HBitmap;
begin
 (* Winkel von -180 bis 180 *)
 if (Angle<-180) then Angle:=-180;
 if (Angle>180) then Angle:=180;
 (* Spezialflle abfangen und optimieren *)
 if (Angle=0) then begin
  LinearBlendFill(DC,Rgn,Rect,Col1,Col2,lbfVert,ColBands);
  Exit;
 end;
 if (Abs(Angle)=180) then begin
  LinearBlendFill(DC,Rgn,Rect,Col2,Col1,lbfVert,ColBands);
  Exit;
 end;
 if (Angle=90) then begin
  LinearBlendFill(DC,Rgn,Rect,Col2,Col1,lbfHorz,ColBands);
  Exit;
 end;
 if (Angle=-90) then begin
  LinearBlendFill(DC,Rgn,Rect,Col1,Col2,lbfHorz,ColBands);
  Exit;
 end;
 (* initialisieren *)
 SelectClipRgn(DC,Rgn);
 RClip:=Rect;
 MemDC:=CreateCompatibleDC(DC);
 Bitmap:=SelectObject(MemDC,CreateCompatibleBitmap(DC,RClip.right,RClip.bottom));
 ColBands:=CorrectColBands(ColBands);
 PrepareColors(Col1,Col2,ColBands,BlendCol);
 (* Startpunkt setzen, Lnge berechnen, Unterscheidung in vier Bereiche *)
 mx:=0; my:=0; d:=0;
 if (Angle>0)and(Angle<90) then begin
  mx:=RClip.Left; my:=RClip.Top+RClip.Bottom;
  d:=sin(Rad(90-Angle))*RClip.Right+cos(Rad(90-Angle))*RClip.Bottom;
 end;
 if (Angle>90)and(Angle<180) then begin
  mx:=RClip.Left+RClip.Right; my:=RClip.Top+RClip.Bottom;
  d:=sin(Rad(Angle-90))*RClip.Right+cos(Rad(Angle-90))*RClip.Bottom;
 end;
 if (Angle<0)and(Angle>-90) then begin
  mx:=RClip.Left; my:=RClip.Top;
  d:=sin(Rad(Angle+90))*RClip.Right+cos(Rad(Angle+90))*RClip.Bottom;
 end;
 if (Angle<-90)and(Angle>-180) then begin
  mx:=RClip.Left+RClip.Right; my:=RClip.Top;
  d:=sin(Rad(-90-Angle))*RClip.Right+cos(Rad(-90-Angle))*RClip.Bottom;
 end;
 Start0.X:=mx+Round(Cos(-Rad(Angle)-Pi/2)*(RClip.Right+RClip.Bottom));
 Start0.Y:=my+Round(Sin(-Rad(Angle)-Pi/2)*(RClip.Right+RClip.Bottom));
 Start1.X:=mx+Round(Cos(-Rad(Angle)+Pi/2)*(RClip.Right+RClip.Bottom));
 Start1.Y:=my+Round(Sin(-Rad(Angle)+Pi/2)*(RClip.Right+RClip.Bottom));
 (* Schrittweite berechnen *)
 sx:=Cos(-Rad(Angle))*d/ColBands;
 sy:=Sin(-Rad(Angle))*d/ColBands;
 Pen:=SelectObject(MemDC,GetStockObject(Null_Pen));
 for z:=1 to ColBands do begin
  Pat[0].X:=Start1.X+Round((z-1)*sx);
  Pat[0].Y:=Start1.Y+Round((z-1)*sy);
  Pat[1].X:=Start0.X+Round((z-1)*sx);
  Pat[1].Y:=Start0.Y+Round((z-1)*sy);
  Pat[2].X:=Start0.X+Round(z*sx);
  Pat[2].Y:=Start0.Y+Round(z*sy);
  Pat[3].X:=Start1.X+Round(z*sx);
  Pat[3].Y:=Start1.Y+Round(z*sy);
  Brush:=SelectObject(MemDC,CreateSolidBrush(BlendCol[z]));
  Polygon(MemDC,Pat,4);
  DeleteObject(SelectObject(MemDC,Brush));
 end;
 DeleteObject(SelectObject(MemDC,Pen));
 SelectClipRgn(DC,Rgn);
 BitBlt(DC,RClip.left,RClip.top,RClip.right,RClip.bottom,MemDC,RClip.left,RClip.top,SrcCopy);
 SelectClipRgn(DC,0);
 DeleteObject(SelectObject(MemDC,Bitmap));
 DeleteDC(MemDC);
end;

(* 4 Farben *)
procedure QuadBlendFill;
var
 RClip,RPnt  : TRect;
 x,y         : Integer;
 Brush       : HBrush;
 Col         : TColor;
 C1,C2,C3,C4 : array[1..3]of Real;
 Dif12,Dif43 : array[1..3]of Real;
 BDif12      : array[1..3]of Real;
 BCol1,BCol2 : array[1..3]of Real;
 w,h,wx,hy   : Real;
 MemDC       : HDC;
 Bitmap      : HBitmap;
begin
 (* Initialisieren *)
 RClip:=Rect;
 MemDC:=CreateCompatibleDC(DC);
 Bitmap:=SelectObject(MemDC,CreateCompatibleBitmap(DC,RClip.right,RClip.bottom));
 if (xTile<=0) then xTile:=1;
 if (xTile>RClip.right) then xTile:=RClip.right;
 if (yTile<=0) then yTile:=1;
 if (yTile>RClip.bottom) then yTile:=RClip.bottom;
 (* Typkonvertierungen wegen Rundungsfehlern *)
 w:=RClip.right;
 h:=RClip.bottom;
 C1[1]:=GetRValue(Col1); C1[2]:=GetGValue(Col1); C1[3]:=GetBValue(Col1);
 C2[1]:=GetRValue(Col2); C2[2]:=GetGValue(Col2); C2[3]:=GetBValue(Col2);
 C3[1]:=GetRValue(Col3); C3[2]:=GetGValue(Col3); C3[3]:=GetBValue(Col3);
 C4[1]:=GetRValue(Col4); C4[2]:=GetGValue(Col4); C4[3]:=GetBValue(Col4);
 RPnt.right:=Trunc(w/xTile)+1;
 RPnt.bottom:=Trunc(h/yTile)+1;
 (* Brechnungen aus der Schleife ziehen *)
 Dif12[1]:=(C1[1]-C2[1])/xTile; Dif12[2]:=(C1[2]-C2[2])/xTile;
 Dif12[3]:=(C1[3]-C2[3])/xTile;
 Dif43[1]:=(C4[1]-C3[1])/xTile; Dif43[2]:=(C4[2]-C3[2])/xTile;
 Dif43[3]:=(C4[3]-C3[3])/xTile;
 wx:=w/xTile;
 hy:=h/yTile;
 for x:=0 to (xTile-1) do begin
  BCol1[1]:=C1[1]-x*Dif12[1]; BCol1[2]:=C1[2]-x*Dif12[2];
  BCol1[3]:=C1[3]-x*Dif12[3];
  BCol2[1]:=C4[1]-x*Dif43[1]; BCol2[2]:=C4[2]-x*Dif43[2];
  BCol2[3]:=C4[3]-x*Dif43[3];
  BDif12[1]:=(BCol1[1]-BCol2[1])/yTile; BDif12[2]:=(BCol1[2]-BCol2[2])/yTile;
  BDif12[3]:=(BCol1[3]-BCol2[3])/yTile;
  for y:=0 to (yTile-1) do begin
   RPnt.left:=RClip.left+Trunc(x*wx);
   RPnt.top:=RClip.top+Trunc(y*hy);
   (* entgltige Farbe berechnen *)
   Col:=RGB(Trunc(BCol1[1]-y*BDif12[1]),Round(BCol1[2]-y*BDif12[2]),Round(BCol1[3]-y*BDif12[3]));
   Brush:=SelectObject(MemDC,CreateSolidBrush(Col));
   PatBlt(MemDC,RPnt.Left,RPnt.Top,RPnt.Right,RPnt.Bottom,PatCopy);
   DeleteObject(SelectObject(MemDC,Brush));
  end;
 end;
 SelectClipRgn(DC,Rgn);
 BitBlt(DC,RClip.left,RClip.top,RClip.right,RClip.bottom,MemDC,RClip.left,RClip.top,SrcCopy);
 SelectClipRgn(DC,0);
 DeleteObject(SelectObject(MemDC,Bitmap));
 DeleteDC(MemDC);
end;

end.
