unit MouseScroolControl;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, typinfo, math, extctrls;

type
  TMouseScroolControl = class(TComponent)
  private
    FControlNames: TStrings;
    procedure SetControlNames(const Value: TStrings);
  protected
    Procedure Loaded; Override;
  public
    Constructor Create(aowner: TComponent); override;
    destructor destroy; override;
  published
    Property ControlNames: TStrings read FControlNames write SetControlNames;
  end;

  TSizeAndScroolPanel = class (TPanel)
  private
    FOnPaint: TNotifyEvent;
  Protected
    fx, fy, fsx, fsy: integer;
    FMouseState: Boolean;
    p : TPanel;
    panelResize: TPoint;
    Procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    Procedure OnPMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    Procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    Procedure OnPMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    Procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    Procedure OnPMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    Procedure Resize; override;
    procedure paint; override;
  Public
    Constructor Create(AOwner: TComponent); override;
    Destructor Destroy; override;
    property canvas;
  published
    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('CDBUtil', [TMouseScroolControl, TSizeAndScroolPanel]);    //wgg was exemples
end;

Type
  TMagicControl = class (TControl)
  Published
    Property OnMouseMove;
    Property OnMouseUp;
    Property OnMouseDown;
  end;

{ TMouseScroolControl }

type
  TScroolControlDatas=  class
  Private
    fx, fy, fsx, fsy: integer;
    FMouseState: Boolean;
    FMouseUp, FMouseDown: TMouseEvent;
    FMouseMove: TMouseMoveEvent;
    FControl: TControl;
  protected
    Procedure MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); virtual;
    Procedure MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); virtual;
    procedure MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer); virtual;
  public
    Constructor Create(Control: TControl);
    property Component: TControl read FControl;
  end;

constructor TMouseScroolControl.Create(aowner: TComponent);
begin
  inherited create(AOwner);
  FControlNames:= TStringList.Create;
end;

destructor TMouseScroolControl.destroy;
begin
  FControlNames.Free;
  inherited destroy;
end;

procedure TMouseScroolControl.Loaded;
  Procedure UpdateControls(c: TComponent);
  var
    j, i: integer;
    ct: TControl;
  begin
    if (c<>nil) and (c is TWinControl) then
      with TWinControl(c) do
        for j := 0 to ControlCount-1 do    // Iterate
        begin
          ct:= Controls[j];
          i:= FControlNames.IndexOf(ct.name);
          if i<>-1 then
            FControlNames.Objects[i]:= TScroolControlDatas.Create(ct);
          if ct is TWinControl then
            UpdateControls(ct as TWinControl);
        end;
  end;
var
  i: integer;
begin
  UpdateControls(Owner);
  if (Owner<>nil) and (Owner is TControl) then
  Begin
    i:= FControlNames.IndexOf(Owner.name);
    if i<>-1 then
      FControlNames.Objects[i]:= TScroolControlDatas.Create(TControl(Owner));
  end;
end;

procedure TMouseScroolControl.SetControlNames(const Value: TStrings);
begin
  FControlNames.assign(Value);
end;

  { TScroolControlDatas }

Constructor TScroolControlDatas.Create(Control: TControl);
Begin
  Inherited create;
  FMouseState:= False;
  FControl:= Control;

  FMouseMove:= TMagicControl(Control).OnMouseMove;
  FMouseUp:= TMagicControl(Control).OnMouseUp;
  FMouseDown:= TMagicControl(Control).OnMouseDown;

  TMagicControl(Control).OnMouseMove:= MouseMove;
  TMagicControl(Control).OnMouseUp:= MouseUp;
  TMagicControl(Control).OnMouseDown:= MouseDown;

end;

procedure TScroolControlDatas.MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  case Button of
    mbRight:
      if FMouseState then
      Begin
        FMouseState:= False;
        FControl.top:= Fsy;
        FControl.Left:= Fsx;
      end;
    mbLeft:
      Begin
        FMouseState:= True;
        Fsx:= FControl.Left;
        Fsy:= FControl.top;
        fx:= fsx+x;
        fy:= fsy+y;
      end;
  end;
  if assigned(FMouseDown) then
    FMouseDown(Sender, Button, Shift, x, y);
end;

procedure TScroolControlDatas.MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
  tx, ty: integer;
  xx, yy: integer;
begin
  if FMouseState then
  Begin
    x:= FControl.left+x;
    y:= FControl.top+y;
    tx:= x-fx;
    ty:= y-fy;
    fx:= x;
    fy:= y;
    xx:= max(FControl.Left+tx,0);
    yy:= max(FControl.top+ty,0);
    if FControl.Parent<>nil then
    Begin
      xx:= min(xx, FControl.parent.clientWidth-FControl.width);
      yy:= min(yy, FControl.parent.clientHeight-FControl.Height);
    end;
    FControl.SetBounds(xx, yy, FControl.Width, FControl.Height);
  end;
  if assigned(FMouseMove) then
    FMouseMove(Sender, Shift, x, y);
end;

procedure TScroolControlDatas.MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button=mbLeft then
    FMouseState:= False;
  if assigned(FMouseUp) then
    FMouseUp(Sender, Button, Shift, x, y);
end;

{ TSizeAndScroolPanel }

constructor TSizeAndScroolPanel.Create(AOwner: TComponent);
begin
  inherited;
  panelResize.X:= -1;
  panelResize.X:= -1;
  p:= TPanel.Create(Self);
  p.Parent:= Self;
  p.Color:= ClBlack;
  p.OnMouseDown:= OnPMouseDown;
  p.OnMouseUp:= OnPMouseUp;
  p.OnMouseMove:= OnPMouseMove;
  SetBounds(left, top, 40, 30);
  Constraints.MinHeight:= 10;
  Constraints.MinWidth:= 10;
end;

destructor TSizeAndScroolPanel.Destroy;
begin
  p.Free;
  inherited;
end;

procedure TSizeAndScroolPanel.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  case Button of
    mbRight:
      if FMouseState then
      Begin
        FMouseState:= False;
        top:= Fsy;
        Left:= Fsx;
      end;
    mbLeft:
      Begin
        if (ssShift in Shift) then   // wgg we require shift to avoid false drags
        begin
          FMouseState:= True;
          Fsx:= Left;
          Fsy:= top;
          fx:= fsx+x;
          fy:= fsy+y;
          Screen.Cursor:= crSizeAll;           // wgg
        end;
      end;
  end;
  BringToFront;
  inherited;
end;

procedure TSizeAndScroolPanel.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  tx, ty: integer;
  xx, yy: integer;
begin
  if FMouseState then
  Begin
    x:= left+x;
    y:= top+y;
    tx:= x-fx;
    ty:= y-fy;
    fx:= x;
    fy:= y;
    xx:= max(Left+tx,0);
    yy:= max(top+ty,0);
    if Parent<>nil then
    Begin
      xx:= min(xx, parent.clientWidth-width);
      yy:= min(yy, parent.clientHeight-Height);
    end;
    SetBounds(xx, yy, Width, Height);
  end;
  Inherited;
end;

procedure TSizeAndScroolPanel.MouseUp(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Button=mbLeft then
    FMouseState:= False;
  Screen.Cursor:= crDefault;                             // wgg
  inherited;
end;

procedure TSizeAndScroolPanel.OnPMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if (Button = mbLeft) And (ssShift in Shift) then    // wgg we require shift to avoid false drags
  Begin
    panelResize.x:= p.left+x;
    panelResize.y:= p.top+y;
    Screen.Cursor:= crSizeAll;           // wgg
  end;
  BringToFront;
end;

procedure TSizeAndScroolPanel.OnPMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  t: TPoint;
begin
  if PanelResize.x<>-1 then
  Begin
    t:= PanelResize;
    t.x:= max(p.left+x, Constraints.MinWidth-9)-t.x;
    t.y:= max(p.top+y, Constraints.MinHeight-9)-t.y;
    inc(PanelResize.x, t.x);
    inc(PanelResize.y, t.y);
    t.x:= max(p.left+t.x, Constraints.MinWidth-9);
    t.y:= max(p.top+t.y, Constraints.MinHeight-9);
//    p.SetBounds(t.x, t.y, 8, 8);
//    p.Parent.Width:= p.left+8;
//    p.Parent.Height:= p.top+8;
    SetBounds(left, top, t.X+9, t.y+9);
  end;
end;

procedure TSizeAndScroolPanel.OnPMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  panelResize.x:= -1;
  panelResize.y:= -1;
  Screen.Cursor:= crDefault;                             // wgg
end;

procedure TSizeAndScroolPanel.paint;
begin
  inherited;
  if (p.top<>Height-9) or (p.left<>Width-9) or (p.width<>8) or (p.height<>8) then
    resize;
  if assigned(FOnPaint) then
    FOnPaint(self);
end;

procedure TSizeAndScroolPanel.Resize;
var
  x, y: integer;
  b: boolean;
begin
  if Parent<>nil then
  Begin
    b:= false;
    x:= width;
    if Left+width>Parent.ClientWidth then
    Begin
      x:= Parent.ClientWidth-Left;
      b:= true;
    end;
    y:= height;
    if top+Height>Parent.ClientHeight then
    begin
      y:= Parent.ClientHeight-top;
      b:= true;
    end;
    if b then
      SetBounds(left, top, x, y);
  end;
  p.SetBounds(Width-9, Height-9, 8, 8);
  inherited;
end;

end.
