unit Mandel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, dib, extctrls, stdctrls;

type
  TMandel = class(TCustomPanel)
  protected
    mx1, my1, mx2, my2 : double;
    cx1, cx2, cy1, cy2 : integer;
    dib1 : TDib8;
    ProcSpd : integer;
    spdCalc : string;
    FOnPaint : TNotifyEvent;
    FCantPaint : boolean;
    FSelect : boolean;
    FOnRecalc : TNotifyEvent;
    Procedure SetDouble(index : integer; v:double);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUP(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure SetProcSpd(v: integer);
    procedure recalc;
    Procedure Loaded; override;
  public
    procedure DoDefaultMouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); virtual;
    procedure DoDefaultMouseUP(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); virtual;
    Procedure DoDefaultMouseMove(Shift: TShiftState; X, Y: Integer); virtual;
    constructor create(aowner : tcomponent); override;
    destructor destroy; override;
    procedure Paint; override;
    procedure Resize; override;
    Procedure XYToFXY(x,y:integer; var fx,fy:double); virtual;
    Procedure Zoom(a:double);
    Procedure Reset;
    property Select : boolean read FSelect;
  published
    property X1 : double index 0 read mx1 write SetDouble;
    property Y1 : double index 1 read my1 write SetDouble;
    property X2 : double index 2 read mx2 write SetDouble;
    property Y2 : double index 3 read my2 write SetDouble;
    property ProcesseurSpd : integer read ProcSpd write SetProcSpd;
    property SpdCalcul : string read SpdCalc write SpdCalc;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    property width;
    property height;
    property top;
    property left;
    property OnPaint : TNotifyEvent read FOnPaint write FOnPaint;
    property Align;
    property ShowHint;
    property Visible;
    property OnRecalc : TNotifyEvent read FOnRecalc write FOnRecalc;
  end;

  TMandelPannel = class(TCustomPanel)
  private
  protected
    m: TMandel;
    l: TLabel;
    b: array [0..2] of TButton;
    procedure BC(Sender: TObject);
    procedure rc(Sender: TObject);
  public
    constructor create(aowner: TComponent); override;
    destructor destroy; override;
  published
  end;

procedure Register;

implementation

const
  FDesignBcle= 10;

Procedure TMandel.Reset;
begin
  FSelect := false;
  mx1:=-2;
  mx2:=1;
  my1:=-1;
  my2:=1;
  cx1:=-1;
  cy1:=-1;
  cx2:=-1;
  cy2:=-1;
  Recalc;
end;

constructor TMandel.create(aowner : tcomponent);
var
  I: int64;
  j: integer;
begin
  inherited create(aowner);
  parent:= nil;
  dib1:= TDib8.Create;
  dib1.width:= 320;
  dib1.Height:= 200;
  width:= 320;
  Height:= 200;
  top:= 0;
  left:= 0;
  QueryPerformanceFrequency(i);
  ProcSpd:=i;
  if not (csDesigning in ComponentState) then
  begin
    for j:=10 to 110 do
      if j<20 then
        dib1.colors[j]:=((j-10)*25)
      else
        if j<60 then
          dib1.colors[j]:=((j-20)*6) shl 8
        else
          dib1.colors[j]:=((j-60)*5) shl 16;
  end else begin
    for j:=10 to FDesignBcle+10 do
      dib1.colors[j]:=(j-10)*(235 div FDesignBcle);
  end;
  Reset;
end;

procedure TMandel.Paint;
begin
  if not assigned(parent) then
    exit;
  dib1.AssociationPal(Canvas.handle);
  dib1.DrawToDC(Canvas.handle,0,0);
  if Assigned(FOnPaint) then
    FOnPaint(self);
end;

destructor TMandel.destroy;
begin
  dib1.free;
  inherited destroy;
end;

procedure TMandel.Resize;
begin
  dib1.width := width and $fffffffc;
  dib1.Height := Height;
  Recalc;
end;

Procedure TMandel.SetDouble(index : integer; v:double);
begin
  case index of    //
    0: mx1 := v;
    1: my1 := v;
    2: mx2 := v;
    3: my2 := v;
  end;    //case
  recalc;
end;

Procedure TMandel.XYToFXY(x,y:integer; var fx,fy:double);
var
  a,b:double;
Begin
  a:=mx2-mx1;
  b:=my2-my1;
  fx:=mx1+a/dib1.width*x;
  fy:=my1+b/dib1.Height*(dib1.Height-y-1);
End;

procedure TMandel.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if assigned(OnMouseDown) then
    inherited MouseDown(button, shift, x, y)
  else
    DoDefaultMouseDown(button, shift, x, y);
end;

procedure TMandel.MouseUP(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if assigned(OnMouseUp) then
    inherited MouseUp(button, shift, x, y)
  else
    DoDefaultMouseUp(button, shift, x, y);
end;

procedure TMandel.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  if assigned(OnMouseMove) then
    inherited MouseMove(Shift, x, y)
  else
    DoDefaultMouseMove(shift, x, y);
end;


procedure TMandel.DoDefaultMouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button=mbLeft then
  begin
    cx1 := x;
    cy1 := y;
    FSelect := True;
  end;
  if Button=mbRight then
  begin
    FSelect := False;
    dib1.DrawToDC(Canvas.handle,0,0);
  end;
end;


Procedure TMandel.DoDefaultMouseUP(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure excg(var a,b:integer);
  var t:integer;
  begin
    t:=a;
    a:=b;
    b:=t;
  end;
var
  a,b:double;
begin
  if FSelect and (Button = MbLeft) then
  begin
    FSelect := false;
    cx2:=x;
    cy2:=y;
    if cx1>cx2 then
      excg(cx1,cx2);
    if cy1<cy2 then
      excg(cy1,cy2);
    XYToFXY(cx1,cy1,a,b);
    XYToFXY(cx2,cy2,mx2,my2);
    mx1:=a;
    my1:=b;
    Recalc;
  end;
end;

Procedure TMandel.DoDefaultMouseMove(Shift: TShiftState; X, Y: Integer);
begin
  if FSelect then
  begin
    dib1.DrawToDC(Canvas.handle,0,0);
    canvas.Brush.Style:=bsClear;
    canvas.Pen.Color:=ClBlack;
    Canvas.Rectangle(cx1,cy1,x,y);
  end
end;

Procedure TMandel.Zoom(a:double);
var
  b,c:double;
begin
  FSelect := false;
  b:=(mx2+mx1)/2;
  c:=(my2+my1)/2;
  mx1:=(mx1-b)/a+b;
  my1:=(my1-c)/a+c;
  mx2:=(mx2-b)/a+b;
  my2:=(my2-c)/a+c;
  Recalc;
end;


procedure Register;
begin
  RegisterComponents('CDBUtil', [TMandel]);                                      //wgg was exemples
  RegisterComponents('CDBUtil', [TMandelPannel]);
end;

procedure TMandel.SetProcSpd(v: integer);
begin

end;

procedure TMandel.recalc;
var
  t, t1: ttime;
begin
  if (Dib1.Height<10) or (Dib1.width<10) then
    exit;
  if csDesigning in ComponentState then
    dib1.mandel2(mx1,my1,mx2,my2,FDesignBcle)
  else
    if not (csLoading in ComponentState) and (not FCantPaint) then
    begin
      assert(setthreadPriority(GetCurrentThread,THREAD_PRIORITY_TIME_CRITICAL));
      t:=now;
      dib1.mandel(mx1,my1,mx2,my2);
      t1:=now;
      assert(setthreadPriority(GetCurrentThread,THREAD_PRIORITY_NORMAL));
      spdCalc:=copy(FloatToStr((t1-t)*3600*24),1,7)+'s';
      if (mx1=-2) and (mx2=1) and (my1=-1) and (my2=1) then
        spdCalc:=SpdCalc+' '+copy(floattostr(ProcSpd/1943345*(t1-t)*3600*24),1,5)+' cycles per loop';
    end;
  if assigned(FOnRecalc) then
    FOnRecalc(self);
  paint;
end;

{ TMandelPannel }

procedure TMandelPannel.BC(Sender: TObject);
begin
  if sender = b[0] then
    m.Reset;
  if sender = b[1] then
    m.Zoom(2);
  if sender = b[2] then
    Application.MessageBox('Mandelbrot drawer by Cyrille de Brebisson.'#13#10' Drawing routine in 100% assembly.'#13#10'All done with Delphi.'#13#10'For informations: cyrille.de-brebisson@capway.com.'#13#10'Use Drag&drop to zoom','About',mb_ok);
end;

constructor TMandelPannel.create(aowner: TComponent);
const
  t: array [0..2] of string = ('Reset', 'Zoom', 'About');
var
  i: integer;
begin
  inherited create(aowner);
  width:= 322;
  height:= 240;
  l:= TLabel.Create(self);
  l.Parent:= self;
  l.top:= 225;
  l.width:= 320;
  l.Left:= 1;
  l.AutoSize:= false;
  l.Alignment:= taCenter;
  l.Font.Size:= l.Font.Size-1;
  m:=TMandel.Create(self);
  m.parent:= self;
  m.top:= 1;
  m.left:=1;
  m.OnRecalc:= rc;
  l.Caption:= m.SpdCalcul;
  for i:=0 to 2 do
  Begin
    b[i]:= TButton.Create(self);
    b[i].Parent:= self;
    b[i].top:= 205;
    b[i].left:= 10 + ((75+37) * i);
    b[i].width:= 75;
    b[i].height:= 15;
    b[i].caption:= t[i];
    b[i].OnClick:= BC;
  end;
end;

destructor TMandelPannel.destroy;
var
  i: integer;
begin
  m.free;
  l.free;
  for i:=0 to 2 do
    b[i].free;
  inherited destroy;
end;

procedure TMandelPannel.rc(Sender: TObject);
begin
  l.Caption:= m.SpdCalcul;
end;

procedure TMandel.Loaded;
begin
  inherited loaded;
  Resize;
end;

end.
