{
  Cyrille de Brebisson utility classes
  This unit is copyright Cyrille de Brebisson
}
unit Utils2;

interface

uses classes, stdctrls, windows, messages;

const
  NbSubList = 127;

Type
  TSpeedStringList = class;

  TSpeedListObject = class
  protected
    FName: String;
    FSpeedList: TSpeedStringList;
    fobject: tobject;
    procedure SetName(const Value: String); virtual;
  public
    Property Name: String read FName write SetName;
    constructor create(name: string);
    destructor destroy; override;
    property SpeedList: TSpeedStringList read FSpeedList write FSpeedList;
    property pointer: tobject read fobject write fobject;
  end;

  PSpeedListObjects = ^TSpeedListObjects;                                       //gp 12/30/1998
  TSpeedListObjects = array [0..0] of TSpeedListObject;                         //gp 12/30/1998

  TSpeedStringList = class
  private
    function GetText: string;
    procedure SetText(const Value: string);
    function GetInObject(Index: Integer): TObject;
    procedure SetInObject(Index: Integer; const Value: TObject);
  Protected
    FOnChange: TNotifyEvent;
    SumOfUsed: array [0..NbSubList-1] of integer;
    datasUsed: array [0..NbSubList-1] of integer;
    datas: array [0..NbSubList-1] of PSpeedListObjects;                         //gp 12/30/1998
    lengthDatas: array [0..NbSubList-1] of integer;                             //gp 12/30/1998
    procedure Changed; virtual;
    function Get(Index: Integer): string; virtual;
    function GetObject(Index: Integer): TSpeedListObject;
    function GetCount: integer;
    Function GetStringList: TStrings;
    Procedure SetStringList(const value: TStrings);
  public
    Procedure NameChange(const obj: TSpeedListObject; const NewName: String);
    Procedure ObjectDeleted(const obj: TSpeedListObject);

    destructor Destroy; override;
    constructor create;
    function Add(const Value: TSpeedListObject): Integer; overload;
    function Add(const Value: String): TSpeedListObject; overload;
    procedure Clear;
    function Find(const name: String): TSpeedListObject;
    function FindPartial(const name: String): TSpeedListObject;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property Objects[Index: Integer]: TSpeedListObject read GetObject;
    property inobject[Index: Integer]: TObject read GetInObject Write SetInObject;
    property Strings[Index: Integer]: string read Get; default;
    property count: integer read GetCount;
    Property StringList: TStrings read GetStringList write SetStringList;
    property text: string read GetText write SetText;
  end;

  TMyList = class(TList)
  protected
    FIndex: Integer;
    FOnChange: TNotifyEvent;
    Procedure Change;
    function Get(Index: Integer): TObject;
    procedure Put(Index: Integer; const Value: TObject);
    Function GetStringRep(o: TObject): string; virtual; 
  public
    destructor Destroy; override;
    function Add(Item: TObject): Integer;
    procedure Clear; override;
    Function Delete(Index: Integer): TObject;
    procedure Exchange(Index1, Index2: Integer);
    function First: TObject;
    function IndexOf(Item: TObject): Integer;
    procedure Insert(Index: Integer; Item: TObject);
    function Last: TObject;
    procedure Move(CurIndex, NewIndex: Integer);
    function Remove(Item: TObject): Integer;
    property Items[Index: Integer]: TObject read Get write Put; default;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    Function IndexOfString(s: string): integer;

    Procedure BOL;
    Function EOL: Boolean;
    Function NextItem: TObject;
    Property CurrentIndex: Integer read FIndex Write FIndex;
  end;

  Procedure LoadComboBox(c: TComboBox; st: TStrings);
  Procedure LoadListBox(c: TListBox; st: TStrings);

implementation

Type
  TLoadThread = class (TThread)
  protected
    c: THandle;
    st: Tstrings;
  Public
    procedure Execute; override;
    Constructor Create(c: TComboBox; st: TStrings);
  end;

{ TLoadThread }

Constructor TLoadThread.Create(c: TComboBox; st: TStrings);
begin
  Self.C:= C.Handle;
  Self.st:= st;
  inherited Create(false);
  FreeOnTerminate:= True;
  Priority:= tpLower;
end;

procedure TLoadThread.Execute;
var
  l1: TStringList;
  i: integer;
begin
  SendMessage(c, CB_RESETCONTENT, 0, 0);
  if st<>nil then
  Begin
    l1:= TStringList.Create;
    l1.Sorted:= true;
    for i :=0 to st.count-1 do
      l1.add(copy(st[i], 1, 128));
    for i :=0 to l1.count-1 do
      SendMessage(c, CB_ADDSTRING, 0, integer(pchar(l1[i])));
    l1.free;
  end;
end;

Procedure LoadComboBox(c: TComboBox; st: TStrings);
begin
  TLoadThread.Create(c, st);
end;

{ TLoadThread }

Type
  TLoadThread2 = class (TThread)
  protected
    c: TListBox;
    st: Tstrings;
  Public
    procedure Execute; override;
    Constructor Create(c: TListBox; st: TStrings);
  end;

Constructor TLoadThread2.Create(c: TListBox; st: TStrings);
begin
  Self.C:= C;
  Self.st:= st;
  inherited Create(false);
  FreeOnTerminate:= True;
  Priority:= tpLower;
end;

procedure TLoadThread2.Execute;
var
  l1: TStringList;
  i: integer;
begin
  SendMessage(c.handle, LB_RESETCONTENT, 0, 0);
  if st<>nil then
  Begin
    l1:= TStringList.Create;
//    l1.Sorted:= true;
    for i :=0 to st.count-1 do
      l1.add(copy(st[i], 1, 128));
    for i :=0 to l1.count-1 do
      SendMessage(c.handle, LB_ADDSTRING, 0, integer(pchar(l1[i])));
    l1.free;
  end;
end;

Procedure LoadListBox(c: TListBox; st: TStrings);
begin
  TLoadThread2.Create(c, st);
end;

type
  TMatrix=array [0..15, 0..15] of integer;
  TArray=array[0..$100-1] of integer;
var
  CrcTable:TMatrix;
  CrcTableArray:TArray absolute CrcTable;

function StringCrc(Const S: string): integer;
var
  i,l,j,k :integer;
  c:^Byte;
begin
  result:=0;
  l:=length(s);
  c:=@s[1];
  for i:=1 to l do
  begin
    j:=c^;
    inc(c);
    k:=(result and $F) shl 4;
    result:= (result shr 4) xor CrcTAbleArray[k+ (j and $F)];
    k:=(result and $F) shl 4;
    result:= (result shr 4) xor CrcTAbleArray[k+ (j shr 4)];
  end;
end;

procedure CrcInitTable;
var
  crc, input: integer;
Begin
  for crc:= 0 to 15 do
    for input:= 0 to 15 do
      CrcTable[crc, input]:= (crc xor input)*$1081
end;

{ TSpeedListObject }

constructor TSpeedListObject.create(name: string);
begin
  inherited create;
  FName:= name;
end;

destructor TSpeedListObject.destroy;
begin
  if FSpeedList<>nil then
    FSpeedList.ObjectDeleted(Self);
  inherited destroy;
end;

procedure TSpeedListObject.SetName(const Value: String);
begin
  FName := Value;
  if FSpeedList<>nil then
    FSpeedList.NameChange(Self, Value);
end;

{ TSpeedStringList }

function TSpeedStringList.Add(const Value: TSpeedListObject): Integer;
var
  crc: integer;
  i: integer;
begin
  crc:= StringCrc(Value.Name) mod High(Datas)+1;
  if DatasUsed[crc]=lengthDatas[crc] then begin                                 //gp 12/30/1998
    ReallocMem(datas[crc], (lengthDatas[crc]*2+1)*SizeOf(datas[1][0]));         //gp 12/30/1998
    lengthDatas[crc] := lengthDatas[crc]*2+1;                                   //gp 12/30/1998
//  SetLength(Datas[crc], Length(Datas[crc])*2+1);                              //gp 12/30/1998
  end;                                                                          //gp 12/30/1998
  Datas[crc][DatasUsed[crc]]:= Value;
  result:= SumOfUsed[crc]+DatasUsed[crc];
  inc(DatasUsed[crc]);
  for i:= crc+1 to High(SumOfUsed) do
    inc(SumOfUsed[i]);
  Value.SpeedList:= Self;
end;

function TSpeedStringList.Add(const Value: String): TSpeedListObject;
begin
  result:= TSpeedListObject.Create(value);
  add(result);
end;

procedure TSpeedStringList.Changed;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TSpeedStringList.Clear;
var
  i, j: integer;
begin
  for i:= low(datas) to high(datas) do
  Begin
    for j:= 0{low(datas[i])} to DatasUsed[i]-1 do                               //gp 12/30/1998
      datas[i][j].free;
    datasUsed[i]:= 0;
//  SetLength(Datas[i], 0);                                                     //gp 12/30/1998
    ReallocMem(datas[i],0);                                                     //gp 12/30/1998                                                   
    lengthDatas[i] := 0;                                                        //gp 12/30/1998
    SumOfUsed[i]:= 0;
  end;
  Changed;
end;

constructor TSpeedStringList.create;
var
  i: integer;
begin
  inherited Create;
  for i:= Low(Datas) to high(datas) do
  Begin
    SumOfUsed[i]:= 0;
    DatasUsed[i]:= 0;
    lengthDatas[i] := 0;                                                        //gp 12/30/1998
    datas[i] := nil;
  end;
end;

destructor TSpeedStringList.Destroy;
begin
  Clear;
  inherited destroy;
end;

function TSpeedStringList.Find(const name: String): TSpeedListObject;
var
  crc: integer;
  i: integer;
begin
  crc:= StringCrc(name) mod High(Datas)+1;
  for i:= 0 to DatasUsed[crc]-1 do
    if Datas[crc][i].name = name then
    Begin
      result:= Datas[crc][i];
      exit;
    end;
  result:= nil;
end;

function TSpeedStringList.FindPartial(const name: String): TSpeedListObject;
var
  i, j: integer;
begin
  for i:= 0 to high(datas) do
    for j:= 0 to DatasUsed[i]-1 do
      if copy(datas[i][j].name,1,length(name))=name then
      Begin
        result:= Datas[i][j];
        exit;
      end;
  result:= nil;
end;

function TSpeedStringList.Get(Index: Integer): string;
var
  i: integer;
begin
  for i:=low(SumOfUsed)+1 to High(SumOfUsed) do
    if Index>SumOfUsed[i] then
    Begin
      result:= Datas[i-1][Index-SumOfUsed[i-1]].name;
      exit;
    end;
  result:= '';
end;

function TSpeedStringList.GetCount: integer;
begin
  result:= SumOfUsed[High(datas)]+DatasUsed[High(Datas)];
end;

function TSpeedStringList.GetInObject(Index: Integer): TObject;
var
  i: integer;
begin
  for i:=low(SumOfUsed)+1 to High(SumOfUsed) do
    if Index>SumOfUSed[i] then
    Begin
      result:= Datas[i-1][Index-SumOfUsed[i-1]].pointer;
      exit;
    end;
  result:= nil;
end;

function TSpeedStringList.GetObject(Index: Integer): TSpeedListObject;
var
  i: integer;
begin
  for i:=low(SumOfUsed)+1 to High(SumOfUsed) do
    if Index>SumOfUSed[i] then
    Begin
      result:= Datas[i-1][Index-SumOfUsed[i-1]];
      exit;
    end;
  result:= nil;
end;

function TSpeedStringList.GetStringList: TStrings;
var
  i, j: integer;
begin
  result:= TStringList.Create;
  for i:= Low(Datas) to High(Datas) do
    for j:= 0{Low(Datas[i])} to DatasUsed[i]-1 do                               //gp 12/30/1998
      result.add(datas[i][j].name);
end;

function TSpeedStringList.GetText: string;
begin
  with StringList do
  Begin
    result:= Text;
    free;
  end;
end;

procedure TSpeedStringList.NameChange(const Obj: TSpeedListObject; const NewName: String);
var
  crc: integer;
  i: integer;
  j: integer;
begin
  crc:= StringCrc(obj.Name) mod High(Datas)+1;
  for i:= 0 to DatasUsed[crc]-1 do
    if Datas[crc][i] = Obj then
    Begin
      for j:= i+1 to DatasUsed[crc]-1 do
        Datas[i-1]:= Datas[i];
      for j:= crc+1 to High(Datas) do
        dec(SumOfUsed[j]);
      if DatasUsed[crc]<lengthDatas[crc] div 2 then begin                       //gp 12/30/1998
        ReallocMem(Datas[crc],DatasUsed[crc]*SizeOf(Datas[crc][0]));            //gp 12/30/1998
        lengthDatas[crc] := DatasUsed[crc];                                     //gp 12/30/1998
//      SetLength(Datas[crc], DatasUsed[crc]);                                  //gp 12/30/1998
      end;
      Add(obj);
      exit;
    end;
end;

procedure TSpeedStringList.ObjectDeleted(const obj: TSpeedListObject);
var
  crc: integer;
  i: integer;
  j: integer;
begin
  crc:= StringCrc(obj.Name) mod High(Datas)+1;
  for i:= 0 to DatasUsed[crc]-1 do
    if Datas[crc][i] = Obj then
    Begin
      for j:= i+1 to DatasUsed[crc]-1 do
        Datas[i-1]:= Datas[i];
      for j:= crc+1 to High(Datas) do
        dec(SumOfUsed[j]);
      Obj.FSpeedList:= nil;
      exit;
    end;
end;

procedure TSpeedStringList.SetInObject(Index: Integer;
  const Value: TObject);
var
  i: integer;
begin
  for i:=low(SumOfUsed)+1 to High(SumOfUsed) do
    if Index>SumOfUSed[i] then
    Begin
      Datas[i-1][Index-SumOfUsed[i-1]].pointer:= value;
      exit;
    end;
end;

procedure TSpeedStringList.SetStringList(const value: TStrings);
var
  i: integer;
begin
  clear;
  for i:= 0 to Value.Count-1 do
    add(TSpeedListObject.Create(value[i]));
end;

procedure TSpeedStringList.SetText(const Value: string);
var
  s: TStrings;
begin
  s:= TStringList.Create;
  try
    s.Text:= Value;
    StringList:= s;
  finally
    s.Free;
  end;
end;


{ TMyList }

function TMyList.Add(Item: TObject): Integer;
begin
  result:= inherited Add(Item);
  Change;
end;

procedure TMyList.BOL;
begin
  FIndex:= 0;
end;

procedure TMyList.Change;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TMyList.Clear;
begin
  inherited Clear;
  Change;
end;

function TMyList.NextItem: TObject;
begin
  result:= Items[FIndex];
  inc(FIndex);
end;

Function TMyList.Delete(Index: Integer): TObject;
begin
  Result:= Items[Index];
  inherited Delete(Index);
  Change;
end;

destructor TMyList.Destroy;
begin
  Change;
  Inherited Destroy;
end;

function TMyList.EOL: Boolean;
begin
  result:= FIndex=Count;
end;

procedure TMyList.Exchange(Index1, Index2: Integer);
begin
  inherited Exchange(Index1, Index2);
  change;
end;

function TMyList.First: TObject;
begin
  result:= inherited First;
end;

function TMyList.Get(Index: Integer): TObject;
begin
  result:= Inherited Get(Index);
end;

function TMyList.IndexOf(Item: TObject): Integer;
begin
  result:= Inherited IndexOf(Item);
end;

procedure TMyList.Insert(Index: Integer; Item: TObject);
begin
  inherited Insert(Index, Item);
  change;
end;

function TMyList.Last: TObject;
begin
  result:= Inherited Last;
end;

procedure TMyList.Move(CurIndex, NewIndex: Integer);
begin
  Inherited Move(CurIndex, NewIndex);
  Change;
end;

procedure TMyList.Put(Index: Integer; const Value: TObject);
begin
  inherited Put(Index, Value);
  Change;
end;

function TMyList.Remove(Item: TObject): Integer;
begin
  result:= inherited Remove(Item);
  Change;
end;

function TMyList.IndexOfString(s: string): integer;
begin
  result:= 0;
  while (result<count) and (GetStringRep(Items[result])<>s) do
    inc(result);
  if result>=count then
    result:= -1;
end;

function TMyList.GetStringRep(o: TObject): string;
begin
  result:= '';
end;

initialization
  CrcInitTable;
end.
